use std::fs::File;
use std::io::{prelude::*, BufReader};
use std::path::PathBuf;

use anyhow::{Context, Result};
use clap::Parser;

#[derive(Parser, Debug)]
#[clap(author, version)]
/// diff binary files
///
/// bytediff checks whether binary files match, and prints out
/// a list of differences at the byte level if they don't.
struct Cli {
    /// Print out a full hexdump of differing bytes, not just a summary of differences
    #[arg(short, long)]
    detailed: bool,

    /// First file to diff
    #[arg(required = true)]
    file1: PathBuf,

    /// Second file to diff
    #[arg(required = true)]
    file2: PathBuf,
}

#[derive(PartialEq, Eq)]
enum DiffState {
    Same,
    DifferNonZero,
    DifferAZero,
    DifferBZero,
}

fn is_byte_vec_zero(v: &[u8]) -> bool {
    v.iter().all(|c| *c == 0u8)
}

fn compare_byte_vecs(a: &[u8], b: &[u8]) -> DiffState {
    if a == b {
        DiffState::Same
    } else if is_byte_vec_zero(a) {
        DiffState::DifferAZero
    } else if is_byte_vec_zero(b) {
        DiffState::DifferBZero
    } else {
        DiffState::DifferNonZero
    }
}

struct ZipLonger<A: Iterator, B: Iterator> {
    a: A,
    b: B,
}

impl<A: Iterator, B: Iterator> Iterator for ZipLonger<A, B> {
    type Item = (Option<A::Item>, Option<B::Item>);

    fn next(&mut self) -> Option<Self::Item> {
        let a_next = self.a.next();
        let b_next = self.b.next();
        if a_next.is_none() && b_next.is_none() {
            return None;
        }
        Some((a_next, b_next))
    }
}

impl<A: Iterator, B: Iterator> ZipLonger<A, B> {
    fn new(a: A, b: B) -> Self {
        Self { a, b }
    }
}

struct Chunks<I: Iterator> {
    n: usize,
    inner: I,
}

impl<T, F, I: Iterator<Item = Result<T, F>>> Iterator for Chunks<I> {
    type Item = Result<Vec<T>, F>;

    fn next(&mut self) -> Option<Self::Item> {
        let mut values = Vec::new();
        for _ in 0..(self.n) {
            let next = self.inner.next();
            match next {
                Some(Ok(v)) => {
                    values.push(v);
                }
                Some(Err(e)) => {
                    return Some(Err(e));
                }
                None => {
                    break;
                }
            }
        }
        if values.is_empty() {
            return None;
        }
        Some(Ok(values))
    }
}

impl<I: Iterator> Chunks<I> {
    fn new(inner: I, n: usize) -> Self {
        Self { inner, n }
    }
}

fn color_byte_display(this: &[u8], that: &[u8], color: u8) -> String {
    let mut result = String::new();

    for (i, b) in this.iter().enumerate() {
        if i % 2 == 0 && i > 0 {
            result.push(' ');
        }

        if that.get(i) == Some(b) {
            result.push_str(&format!("{:02x}", b));
        } else {
            result.push_str(&format!("\x1b[1;{}m{:02x}\x1b[0m", color, b));
        }
    }

    result
}

fn main() -> Result<()> {
    let cli = Cli::parse();

    let a = cli.file1;
    let b = cli.file2;

    let a_bytes = BufReader::new(
        File::open(a.clone())
            .with_context(|| format!("Failed to open input file {}", a.display()))?,
    )
    .bytes();
    let b_bytes = BufReader::new(
        File::open(b.clone())
            .with_context(|| format!("Failed to open input file {}", b.display()))?,
    )
    .bytes();

    let mut diff_state = DiffState::Same;
    let mut region_start: usize = 0;
    let mut differing_chunks = 0;
    let mut total_chunks = 0;

    for (i, (rva, rvb)) in
        ZipLonger::new(Chunks::new(a_bytes, 16), Chunks::new(b_bytes, 16)).enumerate()
    {
        let va: Vec<u8> = rva
            .transpose()
            .with_context(|| format!("Error reading from input file {}", a.display()))?
            .unwrap_or_default();
        let vb: Vec<u8> = rvb
            .transpose()
            .with_context(|| format!("Error reading from input file {}", b.display()))?
            .unwrap_or_default();

        let new_diff_state = compare_byte_vecs(&va, &vb);

        if cli.detailed {
            if new_diff_state == DiffState::Same {
                if diff_state != DiffState::Same {
                    println!("---");
                }
            } else {
                println!(
                    "{0:08x}: {1}\n{0:08x}: {2}",
                    i * 16,
                    color_byte_display(&va, &vb, 31),
                    color_byte_display(&vb, &va, 32),
                );
            }
        } else {
            if new_diff_state != diff_state {
                if diff_state != DiffState::Same {
                    print!("{0:08x} to {1:08x}: ", region_start, i * 16);
                }
                match diff_state {
                    DiffState::Same => (),
                    DiffState::DifferAZero => {
                        println!("First file is zero bytes");
                    }
                    DiffState::DifferBZero => {
                        println!("Second file is zero bytes");
                    }
                    DiffState::DifferNonZero => {
                        println!("Files differ");
                    }
                }
                region_start = i * 16;
            }
        }
        diff_state = new_diff_state;
        if diff_state != DiffState::Same {
            differing_chunks += 1;
        }
        total_chunks += 1;
    }

    match diff_state {
        DiffState::Same => (),
        DiffState::DifferAZero => {
            println!("First file is zero bytes from {0:08x} to end", region_start);
        }
        DiffState::DifferBZero => {
            println!(
                "Second file is zero bytes from {0:08x} to end",
                region_start
            );
        }
        DiffState::DifferNonZero => {
            println!("Files differ from {0:08x} to end", region_start);
        }
    }

    println!("{}/{} chunks differ ({:.2}%)", differing_chunks, total_chunks, 100.0 * (differing_chunks as f32) / (total_chunks as f32));

    Ok(())
}
